package de.ozgcloud.eingang.xta.zip;

import static de.ozgcloud.eingang.xta.zip.TestZipFileFactory.*;
import static de.ozgcloud.eingang.xta.zip.ZipFileExtractor.*;
import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.io.File;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.util.List;
import java.util.Map;
import java.util.stream.IntStream;
import java.util.zip.ZipEntry;

import org.apache.commons.lang3.StringUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.util.MimeTypeUtils;

import de.ozgcloud.eingang.common.errorhandling.TechnicalException;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import lombok.SneakyThrows;

class ZipFileExtractorTest {

	@Spy
	@InjectMocks
	private ZipFileExtractor extractor;

	@DisplayName("extract incoming files safely")
	@Nested
	class TestExtractIncomingFilesWithSizeLimit {
		@Mock
		IncomingFile incomingZipFile;

		@Mock
		File zipFile;

		@Mock
		IncomingFile outIncomingFile;

		List<IncomingFile> outIncomingFiles;

		@BeforeEach
		void mock() {
			outIncomingFiles = List.of(outIncomingFile);

			when(incomingZipFile.getFile()).thenReturn(zipFile);
			doNothing().when(extractor).verifyLimits(zipFile);

			doReturn(outIncomingFiles).when(extractor).extractIncomingFiles(zipFile);
		}

		@DisplayName("should call verify limits")
		@Test
		void shouldCallVerifySizeLimit() {
			extractor.extractIncomingFilesSafely(incomingZipFile);

			verify(extractor).verifyLimits(zipFile);
		}

		@DisplayName("should return")
		@Test
		void shouldReturn() {
			var output = extractor.extractIncomingFilesSafely(incomingZipFile);

			assertThat(output).isEqualTo(outIncomingFiles);
		}
	}

	@DisplayName("verify limits")
	@Nested
	class TestVerifyLimits {
		@Mock
		File zipFile;

		@DisplayName("should return")
		@Test
		void shouldReturn() {
			when(zipFile.length()).thenReturn((long) ZIP_MAX_TOTAL_SIZE / 2);
			doReturn((long) ZIP_MAX_TOTAL_SIZE).when(extractor).sumUncompressedEntrySizes(zipFile);

			extractor.verifyLimits(zipFile);
		}

		@DisplayName("should throw if size limit exceeded")
		@Test
		void shouldThrowIfSizeLimitExceeded() {
			doReturn((long) ZIP_MAX_TOTAL_SIZE + 1).when(extractor).sumUncompressedEntrySizes(zipFile);

			assertThatThrownBy(() -> extractor.verifyLimits(zipFile))
					.isInstanceOf(TechnicalException.class);
		}

		@DisplayName("should throw if ratio exceeded")
		@Test
		void shouldThrowIfRatioExceeded() {
			when(zipFile.length()).thenReturn(1L);
			doReturn((long) ZIP_MAX_THRESHOLD + 1).when(extractor).sumUncompressedEntrySizes(zipFile);

			assertThatThrownBy(() -> extractor.verifyLimits(zipFile))
					.isInstanceOf(TechnicalException.class);
		}
	}

	@DisplayName("extract incoming files")
	@Nested
	class TestExtractIncomingFiles {

		private File zipFile;

		@BeforeEach
		void mock() {
			zipFile = createTempZipFile(fromMap(Map.of(
					"file1.pdf", "file content1",
					"file2.xml", "<root></root>",
					"file3.png", ""
			)));
		}

		@DisplayName("should contain content")
		@Test
		void shouldContainContent() {
			var extractedFiles = extractor.extractIncomingFiles(zipFile);

			var contents = extractedFiles.stream().map(f -> {
				try {
					return Files.readString(f.getFile().toPath());
				} catch (IOException e) {
					throw new RuntimeException(e);
				}
			}).toList();
			assertThat(contents).containsExactlyInAnyOrder("file content1", "<root></root>", "");
		}

		@DisplayName("should have names")
		@Test
		void shouldHaveNames() {
			var extractedFiles = extractor.extractIncomingFiles(zipFile);

			var names = extractedFiles.stream().map(IncomingFile::getName).toList();
			assertThat(names).containsExactlyInAnyOrder("file1.pdf", "file2.xml", "file3.png");
		}

		@DisplayName("should have content types")
		@Test
		void shouldHaveContentTypes() {
			var extractedFiles = extractor.extractIncomingFiles(zipFile);

			var names = extractedFiles.stream().map(IncomingFile::getContentType).toList();
			assertThat(names).containsExactlyInAnyOrder("application/pdf", "application/xml", "image/png");
		}

		private List<TestZipEntry> fromMap(Map<String, String> entries) {
			return entries.entrySet().stream().map(kv -> TestZipEntry.builder()
							.name(kv.getKey())
							.content(kv.getValue().getBytes(StandardCharsets.UTF_8))
							.build())
					.toList();
		}
	}

	@DisplayName("extract zip bomb")
	@Nested
	class TestExtractZipBomb {
		private static final int SMALLER_MAX_ZIP_FILE_SIZE = 2 * 1024;

		@DisplayName("should throw with too many entries")
		@Test
		void shouldThrow() {
			var zipFile = createIncomingFile(createTempZipFile(IntStream.range(0, ZIP_MAX_ENTRIES + 1).mapToObj(i -> TestZipEntry.builder()
					.name("file%d.txt".formatted(i))
					.content(toBytes("A".repeat(2)))
					.build()
			).toList()));

			assertThatThrownBy(() -> extractor.extractIncomingFilesSafely(zipFile))
					.isInstanceOf(TechnicalException.class);
		}

		@SneakyThrows
		@DisplayName("should throw with fake getSize")
		@Test
		void shouldThrowWithFakeGetSize() {
			doReturn(SMALLER_MAX_ZIP_FILE_SIZE).when(extractor).getZipMaxTotalSize();
			var zipBomb = createIncomingFile(createTempZipBomb(SMALLER_MAX_ZIP_FILE_SIZE));

			assertThatThrownBy(() -> extractor.extractIncomingFilesSafely(zipBomb))
					.isInstanceOf(TechnicalException.class)
					.hasRootCauseMessage(LimitedInputStream.LIMITED_EXCEEDED_MESSAGE);
		}

		@DisplayName("should throw with too large size")
		@Test
		void shouldThrowWithTooLargeSize() {
			doReturn(SMALLER_MAX_ZIP_FILE_SIZE).when(extractor).getZipMaxTotalSize();
			var zipFile = createIncomingFile(createTempZipFile(List.of(TestZipEntry.builder()
					.name("toolargefile.txt")
					.content(toBytes("A".repeat(SMALLER_MAX_ZIP_FILE_SIZE + 1)))
					.build()
			)));

			assertThatThrownBy(() -> extractor.extractIncomingFilesSafely(zipFile))
					.isInstanceOf(TechnicalException.class);
		}

		private IncomingFile createIncomingFile(File file) {
			return IncomingFile.builder()
					.file(file)
					.build();
		}
	}

	@DisplayName("create incoming file")
	@Nested
	class TestCreateIncomingFile {
		@Mock
		File file;

		@Mock
		ZipEntry zipEntry;

		private static final String NAME = "filename.name";
		private static final Long SIZE = 5L;
		private static final String CONTENT_TYPE = "some/content";

		@BeforeEach
		void mock() {
			when(zipEntry.getName()).thenReturn(NAME);
			when(zipEntry.getSize()).thenReturn(SIZE);
			doReturn(CONTENT_TYPE).when(extractor).getContentType(NAME);
		}

		@DisplayName("should have name")
		@Test
		void shouldHaveName() {
			var incomingFile = create();

			assertThat(incomingFile.getName()).isEqualTo(NAME);
		}

		@DisplayName("should have size")
		@Test
		void shouldHaveSize() {
			var incomingFile = create();

			assertThat(incomingFile.getSize()).isEqualTo(SIZE);
		}

		@DisplayName("should have content type")
		@Test
		void shouldHaveContentType() {
			var incomingFile = create();

			assertThat(incomingFile.getContentType()).isEqualTo(CONTENT_TYPE);
		}

		@DisplayName("should have file")
		@Test
		void shouldHaveFile() {
			var incomingFile = create();

			assertThat(incomingFile.getFile()).isEqualTo(file);
		}

		private IncomingFile create() {
			return extractor.createIncomingFile(file, zipEntry);
		}
	}

	@DisplayName("sum uncompressed entry size")
	@Nested
	class TestSumUncompressedEntrySize {
		@DisplayName("should return size")
		@Test
		void shouldReturnSize() {
			var sizes = IntStream.range(100, 110).boxed().toList();
			var expectedSum = sizes.stream().mapToLong(Integer::longValue).sum();
			var zipFile = createTempZipFile(sizes.stream()
					.map(size -> TestZipEntry.builder()
							.name("somefilewithsize%d".formatted(size))
							.content(toBytes("A".repeat(size)))
							.build()
					).toList());

			var sum = extractor.sumUncompressedEntrySizes(zipFile);

			assertThat(sum).isEqualTo(expectedSum);
		}
	}

	@Nested
	class TestContentType {

		@Test
		void shouldReturnDefaultWhenNullString() {
			assertThrows(NullPointerException.class, () -> extractor.getContentType(null));
		}

		@Test
		void shouldReturnDefaultWhenEmptyString() {
			var contentType = extractor.getContentType(StringUtils.EMPTY);

			assertThat(contentType).isEqualTo(MimeTypeUtils.APPLICATION_OCTET_STREAM_VALUE);
		}

		@Test
		void shouldReturnDefaultWhenSpaceString() {
			var contentType = extractor.getContentType(StringUtils.SPACE);

			assertThat(contentType).isEqualTo(MimeTypeUtils.APPLICATION_OCTET_STREAM_VALUE);
		}

		@Test
		void shouldGetContentType() {
			var fileNames = List.of("1.xml", "2.txt");

			var contentTypes = fileNames.stream().map(extractor::getContentType).toList();

			assertThat(contentTypes).containsExactlyInAnyOrder(MimeTypeUtils.APPLICATION_XML_VALUE, MimeTypeUtils.TEXT_PLAIN_VALUE);
		}
	}

	@DisplayName("map zip entries")
	@Nested
	class TestMapZipEntries {

		@DisplayName("should throw if max entries exceeded")
		@Test
		void shouldThrowIfMaxEntriesExceeded() {
			var zipWithTooManyEntries = createTempZipFile(IntStream.range(0, ZIP_MAX_ENTRIES + 1)
					.mapToObj(i -> TestZipEntry.builder()
							.name("test%d.txt".formatted(i))
							.content(toBytes("test file %d".formatted(i)))
							.build()
					).toList());

			assertThatThrownBy(() -> extractor.mapZipEntries(zipWithTooManyEntries, entry -> null))
					.isInstanceOf(TechnicalException.class);
		}

		@DisplayName("should map with mapping function")
		@Test
		void shouldMapWithMappingFunction() {
			var expectedNumberList = IntStream.range(0, ZIP_MAX_ENTRIES).boxed().toList();
			var zipFile = createTempZipFile(expectedNumberList.stream()
					.map(i -> TestZipEntry.builder()
							.name("%d".formatted(i))
							.content(toBytes("some content"))
							.build()
					).toList());

			var numberList = extractor.mapZipEntries(zipFile, entry -> Integer.parseInt(entry.getName()));

			assertThat(numberList).isEqualTo(expectedNumberList);
		}
	}

	private byte[] toBytes(String string) {
		return string.getBytes(StandardCharsets.UTF_8);
	}

}
